
#include <lubishev/Common.h>
#include <lubishev/Exception.h>
#include <lubishev/DBSQLite.h>

namespace lubishev {

DBSQLite::Statement::Statement(sqlite3 *db, const QString &sql) :
        mDB(db),
        mStmt(0) {
}

DBSQLite::Statement::~Statement() {
    finalize();
}

s32 DBSQLite::Statement::columnAsS32(s32 columnID) {
    return sqlite3_column_int(mStmt, columnID);
}

s64 DBSQLite::Statement::columnAsS64(s32 columnID) {
    return sqlite3_column_int64(mStmt, columnID);
}

QString DBSQLite::Statement::columnAsText(s32 columnID) {
    return QString::fromUtf8(
            reinterpret_cast<const char*>(sqlite3_column_text(mStmt, columnID)));
}

void DBSQLite::Statement::setSQL(const QString &sql) {
    finalize();
    if (sqlite3_prepare_v2(mDB, sql.toUtf8().constData(), -1, &mStmt, 0) != SQLITE_OK) {
        finalize();
        LUBISHEV_EXCEPTION("Must not happen",
                           vaToStr("Could not prepare statement '%s': %s",
                                   sql.toUtf8().constData(),
                                   sqlite3_errmsg(mDB)));
    }
    mSQL = sql;
}

// Returns true if there is anything to read.
bool DBSQLite::Statement::step() {
    s32 rc = sqlite3_step(mStmt);
    switch (rc) {
        case SQLITE_ROW:
            return true;
        case SQLITE_DONE:
            return false;
        default:
            LUBISHEV_EXCEPTION(
                "Must not happen",
                vaToStr("Could not process the statement '%s': %s",
                        mSQL.toUtf8().constData(), sqlite3_errmsg(mDB)));
    }
}

void DBSQLite::Statement::finalize() {
    if (mStmt) {
        sqlite3_finalize(mStmt);
        mStmt = 0;
    }
}

DBSQLite::DBSQLite() : mDB(0), mStmt(0) { }

QString DBSQLite::createEvent(time_t start, time_t end) {
    mStmt->setSQL(QString("insert into events values(null, %1, %2)")
                         .arg(start)
                         .arg(end));
    mStmt->step();
    mStmt->setSQL("select max(id) from events");
    mStmt->step();
    return mStmt->columnAsText(0);
}

QStringList DBSQLite::eventTags(const QString &id) {
    mStmt->setSQL(QString("select description "
                          "from events, tags, event_tags "
                          "where events.id = %1 "
                          "  and event_tags.event_id = events.id "
                          "  and event_tags.tag_id   = tags.id").arg(id));
    QStringList list;
    while (mStmt->step())
        list << mStmt->columnAsText(0);
    return list;
}

QStringList DBSQLite::eventsWithin(time_t start, time_t end) {
    mStmt->setSQL(QString("select id from events where "
                          "(start >= %1 and start <= %2) "
                          "or (end >= %1 and end <= %2) "
                          "or (%1 >= start and %1 <= end) "
                          "or (%2 >= start and %2 <= end)").arg(start).arg(end));
    QStringList ids;
    while (mStmt->step())
        ids.push_back(mStmt->columnAsText(0));
    return ids;
}

void DBSQLite::eventTimes(const QString &id, time_t *start, time_t *end) {
    mStmt->setSQL(QString("select start, end from events where id = %1").arg(id));
    mStmt->step();
    if (start)
        *start = mStmt->columnAsS64(0);
    if (end)
        *end = mStmt->columnAsS64(1);
}

void DBSQLite::setDir(const QString &dir) {
    close();
    open(dir + "/lubishev.db");
}

void DBSQLite::setEventTags(const QString &id, const QStringList &list) {
    // Remove current bindings if any.
    mStmt->setSQL(QString("delete from event_tags where event_id = %1").arg(id));
    mStmt->step();
    // Assign new bindings.
    for (s32 i = 0; i < list.size(); i++) {
        mStmt->setSQL(QString("insert into event_tags values(%1, %2)")
                              .arg(id)
                              .arg(tagID(list[i])));
        mStmt->step();
    }
}

void DBSQLite::setEventTimes(const QString &id, time_t start, time_t end) {
    mStmt->setSQL(QString("update events set start = %1, end = %2 where id = %3")
                          .arg(start)
                          .arg(end)
                          .arg(id));
    mStmt->step();
}

void DBSQLite::close() {
    if (mDB) {
        delete mStmt;
        sqlite3_close(mDB);
        mDB = 0;
    }
}

void DBSQLite::ensure() {
    Statement stmt(mDB);
    // Events.
    stmt.setSQL("select * from sqlite_master "
                "where type = 'table' and name = 'events'");
    if (!stmt.step()) {
        stmt.setSQL("create table events(id integer primary key autoincrement, start integer, end integer)");
        stmt.step();
    }
    // Tags.
    stmt.setSQL("select * from sqlite_master "
                "where type = 'table' and name = 'tags'");
    if (!stmt.step()) {
        stmt.setSQL("create table tags(id integer primary key autoincrement, description text)");
        stmt.step();
    }
    // EventTags.
    stmt.setSQL("select * from sqlite_master "
                "where type = 'table' and name = 'event_tags'");
    if (!stmt.step()) {
        stmt.setSQL(
            "create table event_tags(event_id integer, tag_id integer, foreign key(event_id) references events(id))");
        stmt.step();
    }
}

void DBSQLite::open(const QString &fileName) {
    if (sqlite3_open(fileName.toUtf8().constData(), &mDB) != SQLITE_OK) {
        close();
        LUBISHEV_EXCEPTION("Must not happen",
                           vaToStr("Could not open DB file '%s'",
                                   fileName.toUtf8().constData()));
    }
    ensure();
    mStmt = new Statement(mDB);
}

s64 DBSQLite::tagID(const QString &tag) {
    mStmt->setSQL(
        QString("select id from tags where description = '%1'").arg(tag));
    // Retrieve existing one.
    if (mStmt->step())
        return mStmt->columnAsS64(0);
    // Create a new one.
    else {
        mStmt->setSQL(QString("insert into tags values(null, '%1')").arg(tag));
        mStmt->step();
        return tagID(tag);
    }
}

} // namespace lubishev

